#include <stdlib.h>

#include "paramsGPU.h"

#define CUDAGMM_VERSION 5

#if(CUDAGMM_VERSION == 5)

// __constant__ CvFastBgGMMParams d_GMMParams;
// __constant__ CvFastBgGMMData d_GMMData;
// __constant__ int d_arrImageInfo[ImageInfoCount];

/*============================================================================*/
// CUDA-related functions
/*============================================================================*/

/*=======================================================================================*/

__device__ int _cudaUpdateFastBgGMM(int pixel, 
									float red, float green, float blue,
									int* pModesUsed,
				    int * d_arrImageInfo, CvFastBgGMMData d_GMMData,CvFastBgGMMParams d_GMMParams
									)
{
	//calculate distances to the modes (+ sort???)
	//here we need to go in descending order!!!

	int pos;
	bool bFitsPDF = 0;
	int bBackground = 0;
	float m_fOneMinAlpha = 1 - d_GMMParams.fAlphaT;
	int nModes = (*pModesUsed);
	float weight, totalWeight = 0.0f;
	float dR, dG, dB;
	float dist, k, sigmanew;

	//go through all modes
	for (int iModes = 0; iModes < nModes; iModes++)
	{
		pos = pixel + iModes*d_arrImageInfo[ImageInfoPixelCount];
		weight = d_GMMData.rWeight[pos];

		//fit not found yet
		if (!bFitsPDF)
		{
			//check if it belongs to some of the modes
			//calculate distance
			float4 cGauss = d_GMMData.ucGaussian[pos];

			dR = cGauss.x - red;
			dG = cGauss.y - green;
			dB = cGauss.z - blue;

			//check if it fits the current mode (Factor * sigma)

			//square distance -slower and less accurate
			//float maxDistance = cvSqrt(m_fTg*var);
			//if ((fabs(dR) <= maxDistance) && (fabs(dG) <= maxDistance) && (fabs(dB) <= maxDistance))
			//circle
			dist = dR*dR + dG*dG + dB*dB;

			//background? - m_fTb
			if ((totalWeight < d_GMMParams.fTB) && (dist < d_GMMParams.fTb * cGauss.w))
				bBackground = 1;

			//check fit
			if (dist < d_GMMParams.fTg * cGauss.w)
			{
				//belongs to the mode
				bFitsPDF = 1;

				//update distribution
				k = d_GMMParams.fAlphaT/weight;
				weight = m_fOneMinAlpha * weight + d_GMMParams.fPrune;
				weight += d_GMMParams.fAlphaT;
				cGauss.x -= k*(dR);
				cGauss.y -= k*(dG);
				cGauss.z -= k*(dB);

				//limit update speed for cov matrice
				//not needed
				sigmanew = cGauss.w + k*(dist - cGauss.w);

				//limit the variance
				cGauss.w = sigmanew < 4 ? 4 : 
					sigmanew > 5 * d_GMMParams.fSigma ? 5 * d_GMMParams.fSigma : sigmanew;

				d_GMMData.ucGaussian[pos] = cGauss;

				//sort
				//all other weights are at the same place and 
				//only the matched (iModes) is higher -> just find the new place for it

				for (int iLocal = iModes; iLocal > 0; iLocal--)
				{
					int posLocal = pixel + iLocal*d_arrImageInfo[ImageInfoPixelCount];
					if (weight < (d_GMMData.rWeight[posLocal-d_arrImageInfo[ImageInfoPixelCount]]))
					{
						break;
					}
					else
					{
						//swap
						float tmpVal;
						float4 tmpuChar;
						SWAP(d_GMMData.ucGaussian[posLocal],
							d_GMMData.ucGaussian[posLocal - d_arrImageInfo[ImageInfoPixelCount]],
							tmpuChar);
						SWAP(d_GMMData.rWeight[posLocal],
							d_GMMData.rWeight[posLocal - d_arrImageInfo[ImageInfoPixelCount]],
							tmpVal);
					}
				}

				//belongs to the mode
			}
			else
			{
				weight = m_fOneMinAlpha * weight + d_GMMParams.fPrune;

				//check prune
				if (weight < -(d_GMMParams.fPrune))
				{
					weight = 0.0f;
					nModes--;
					//	bPrune=1;
					//break;//the components are sorted so we can skip the rest
				}
			}
			//check if it fits the current mode (2.5 sigma)
			///////
		}	//fit not found yet
		else
		{
			weight = m_fOneMinAlpha * weight + d_GMMParams.fPrune;

			if (weight < -(d_GMMParams.fPrune))
			{
				weight=0.0;
				nModes--;
				//bPrune=1;
				//break;//the components are sorted so we can skip the rest
			}
		}
		totalWeight += weight;
		d_GMMData.rWeight[pos] = weight;
	}
	//go through all modes
	//////

	//renormalize weights
	for (int iLocal = 0; iLocal < nModes; iLocal++)
	{
		d_GMMData.rWeight[pixel + iLocal*d_arrImageInfo[ImageInfoPixelCount]] /= totalWeight;
	}

	//make new mode if needed and exit
	if (!bFitsPDF)
	{
		if (nModes == d_GMMParams.nM)
		{
			//replace the weakest
		}
		else
		{
			//add a new one
			//totalWeight+=m_fAlphaT;
			//pos++;
			nModes++;
		}
		pos = pixel + (nModes-1)*d_arrImageInfo[ImageInfoPixelCount];

		if (nModes == 1)
			d_GMMData.rWeight[pos] = 1;
		else
			d_GMMData.rWeight[pos] = d_GMMParams.fAlphaT;

		//renormalize weights
		for (int iLocal = 0; iLocal < nModes-1; iLocal++)
		{
			d_GMMData.rWeight[pixel + iLocal*d_arrImageInfo[ImageInfoPixelCount]] *= m_fOneMinAlpha;
		}

		float4 cGauss;
		cGauss.x = red;
		cGauss.y = green;
		cGauss.z = blue;
		cGauss.w = d_GMMParams.fSigma;
		d_GMMData.ucGaussian[pos] = cGauss;

		//sort
		//find the new place for it
		for (int iLocal = nModes - 1; iLocal>0; iLocal--)
		{
			int posLocal = pixel + iLocal*d_arrImageInfo[ImageInfoPixelCount];
			if (d_GMMParams.fAlphaT < (d_GMMData.rWeight[posLocal - d_arrImageInfo[ImageInfoPixelCount]]))
			{
				break;
			}
			else
			{
				//swap
				float4 tmpuChar;
				float tmpVal;
				SWAP(d_GMMData.ucGaussian[posLocal],
					d_GMMData.ucGaussian[posLocal - d_arrImageInfo[ImageInfoPixelCount]],
					tmpuChar);
				SWAP(d_GMMData.rWeight[posLocal],
					d_GMMData.rWeight[posLocal - d_arrImageInfo[ImageInfoPixelCount]],
					tmpVal);
			}
		}
	}

	//set the number of modes
	*pModesUsed=nModes;

	return bBackground;
}

/*=======================================================================================*/

/*=======================================================================================*/
__device__ int _cudaRemoveShadowGMM(int pixel, 
									float red, float green, float blue, 
									int nModes,
				    int * d_arrImageInfo, CvFastBgGMMData d_GMMData,CvFastBgGMMParams d_GMMParams)
{
	//calculate distances to the modes (+ sort???)
	//here we need to go in descending order!!!
	//	long posPixel = pixel * m_nM;
	int pos;
	float tWeight = 0;
	float numerator, denominator;

	// check all the distributions, marked as background:
	for (int iModes=0;iModes<nModes;iModes++)
	{
		pos=pixel+iModes*d_arrImageInfo[ImageInfoPixelCount];
		float4 cGauss = d_GMMData.ucGaussian[pos];
		float weight = d_GMMData.rWeight[pos];
		tWeight += weight;

		numerator = red * cGauss.x + green * cGauss.y + blue * cGauss.z;
		denominator = cGauss.x * cGauss.x + cGauss.y * cGauss.y + cGauss.z * cGauss.z;
		// no division by zero allowed
		if (denominator == 0)
		{
			break;
		}
		float a = numerator / denominator;

		// if tau < a < 1 then also check the color distortion
		if ((a <= 1) && (a >= d_GMMParams.fTau))//m_nBeta=1
		{
			float dR=a * cGauss.x - red;
			float dG=a * cGauss.y - green;
			float dB=a * cGauss.z - blue;

			//square distance -slower and less accurate
			//float maxDistance = cvSqrt(m_fTb*var);
			//if ((fabs(dR) <= maxDistance) && (fabs(dG) <= maxDistance) && (fabs(dB) <= maxDistance))
			//circle
			float dist=(dR*dR+dG*dG+dB*dB);
			if (dist<d_GMMParams.fTb*cGauss.w*a*a)
			{
				return 2;
			}
		}
		if (tWeight > d_GMMParams.fTB)
		{
			break;
		}
	}
	return 0;
}

/*=======================================================================================*/

/*=======================================================================================*/

__device__ void _cudaReplacePixelBackgroundGMM(int pixel, uchar4* pData,CvFastBgGMMData d_GMMData  )
{
	uchar4 tmp;
	float4 cGauss = d_GMMData.ucGaussian[pixel];
	tmp.z = (unsigned char) cGauss.x;
	tmp.y = (unsigned char) cGauss.y;
	tmp.x = (unsigned char) cGauss.z;
	(*pData) = tmp;
}
/*=======================================================================================*/

/*=======================================================================================*/
extern __shared__ int sharedInfo[];

template <int BLOCK_SIZE>
__global__ void cudaUpdateFastBgGMM(unsigned char* data, unsigned char* output,
				    int * d_arrImageInfo, CvFastBgGMMData d_GMMData,CvFastBgGMMParams d_GMMParams)
{
	if(threadIdx.x == 0)
	{
		// the start pixel for current block
		sharedInfo[0] = (blockIdx.x * BLOCK_SIZE)*d_arrImageInfo[ImageInfoPixelsPerThread];
	}
	__syncthreads();

	int iPxStart = sharedInfo[0] + threadIdx.x;
	int iPxEnd = min( d_arrImageInfo[ImageInfoPixelCount], 
		sharedInfo[0] + (BLOCK_SIZE * d_arrImageInfo[ImageInfoPixelsPerThread]));

	uchar4* pGlobalInput = ((uchar4*)data) + iPxStart;
	unsigned char* pGlobalOutput = output + iPxStart;

	int* pUsedModes = d_GMMData.rnUsedModes + iPxStart;
	unsigned char fRed, fGreen, fBlue;
	uchar4 currentInputPx;

	for(int i = iPxStart; i < iPxEnd; i += BLOCK_SIZE)
	{
		// retrieves the color
		currentInputPx = *pGlobalInput;
		fBlue = currentInputPx.x;
		fGreen = currentInputPx.y;
		fRed = currentInputPx.z;
		pGlobalInput += BLOCK_SIZE;

		// update model + background subtract
		int result = _cudaUpdateFastBgGMM(i, fRed, fGreen, fBlue, pUsedModes,d_arrImageInfo,d_GMMData,d_GMMParams);
		int nMLocal = *pUsedModes;
		pUsedModes += BLOCK_SIZE;

		if (d_GMMParams.bShadowDetection)
		{
			if (!result)
			{
				result= _cudaRemoveShadowGMM(i, fRed, fGreen, fBlue, nMLocal,d_arrImageInfo,d_GMMData,d_GMMParams);
			}
		}
		switch (result)
		{
		case 0:

			//foreground
			(*pGlobalOutput) = 255;
			if (d_GMMParams.bRemoveForeground) 
			{
				_cudaReplacePixelBackgroundGMM(i, pGlobalInput-BLOCK_SIZE,d_GMMData);
			}
			break;

		case 1:

			//background
			(*pGlobalOutput) = 0;
			break;

		case 2:

			//shadow
			(*pGlobalOutput) = 128;
			if (d_GMMParams.bRemoveForeground) 
			{
				_cudaReplacePixelBackgroundGMM(i, pGlobalInput-BLOCK_SIZE,d_GMMData);
			}

			break;
		}
		pGlobalOutput += BLOCK_SIZE;
	}
}

extern "C" 
void cudaUpdateFastBgGMM_Wrapper(unsigned char* data, unsigned char* output, 
				 int iGridSize, int iBlockSize, int sharedSize, cudaStream_t stream,
				 int * d_arrImageInfo, CvFastBgGMMData d_GMMData,CvFastBgGMMParams d_GMMParams
				)
{
	switch(iBlockSize)
	{
	case 8:
		cudaUpdateFastBgGMM<8> <<< iGridSize, 8, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams );
		break;
	case 16:
		cudaUpdateFastBgGMM<16> <<< iGridSize, 16, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams);
		break;
	case 32:
		cudaUpdateFastBgGMM<32> <<< iGridSize, 32, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams);
		break;
	case 64:
		cudaUpdateFastBgGMM<64> <<< iGridSize, 64, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams );
		break;
	case 128:
		cudaUpdateFastBgGMM<128> <<< iGridSize, 128, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams );
		break;
	case 256:
		cudaUpdateFastBgGMM<256> <<< iGridSize, 256, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams );
		break;
	case 512:
		cudaUpdateFastBgGMM<512> <<< iGridSize, 512, sharedSize, stream>>>
			(data, output,d_arrImageInfo,d_GMMData,d_GMMParams);
		break;
	}
}
#endif
