#include <stdio.h>
#include <stdlib.h>
#include "cv.h"
#include "cvaux.h"
#include "highgui.h"
#include <ml.h>
#include <ctype.h>
//#include "conio.h"
// #include "CvFastBgGMM.h"
#include "params.h"

#include <memory.h>

#include <cuda_runtime.h>
#include <cuda.h>
#include <cutil.h>
#include <cutil_inline.h>




#define TIMER_CUDA	TRUE
#define _DEBUG TRUE
#ifdef _DEBUG
#define	CUDAGMM_SAFE_CALL(x)	CUDA_SAFE_CALL(x)
#else
#define	CUDAGMM_SAFE_CALL(x)	x
#endif

//__constant__ CvFastBgGMMParams d_GMMParams;
//__constant__ CvFastBgGMMData d_GMMData;
// __constant__ int d_arrImageInfo[ImageInfoCount];
//#define TEST_VIDEO "/home/apostolis/sample.avi"

extern "C" 
void cudaUpdateFastBgGMM_Wrapper(unsigned char* data, unsigned char* output, 
				 int iGridSize, int iBlockSize, int sharedSize, cudaStream_t stream,
				 int * d_arrImageInfo, CvFastBgGMMData d_GMMData,CvFastBgGMMParams d_GMMParams 
				);

int InitCUDA(CvFastBgGMM* pGMM)
{




	int count = 0;
	int i = 0;

	cudaGetDeviceCount(&count);
	if(count == 0)
	{
		fprintf(stderr, "There is no device.\n");
		return -1;
	}

	for(i = 0; i < count; i++)
	{
		cudaDeviceProp prop;
		if(cudaGetDeviceProperties(&prop, i) == cudaSuccess)
		{
			if(prop.major >= 1)
			{
				pGMM->nThreadsPerBlock = prop.maxThreadsPerBlock / 4;

				// temporarily hard-code a little here...
				pGMM->nBlocksPerGrid = 256;
				break;
			}
		}
	}
	if(i == count) {
		fprintf(stderr, "There is no device supporting CUDA.\n");
		return -1;
	}
//	cudaSetDevice(i);
//	cutilSafeCall(cudaSetDeviceFlags(cudaDeviceMapHost));
	return i;


}

template <bool toPinned>
void copyImageData(IplImage* h_img, unsigned char* d_pinnedMem, int channels)
{
	if(h_img->widthStep == channels*h_img->width)
	{
		memcpy(
			toPinned ? (d_pinnedMem) : (unsigned char*)(h_img->imageData), 
			toPinned ? (unsigned char*)(h_img->imageData) : (d_pinnedMem),
			h_img->widthStep*h_img->height);
	}
	else
	{
		unsigned char* d_curData = d_pinnedMem;
		if(toPinned)
		{
			for(int i = 0; i < h_img->height; ++i)
			{
				memcpy(
					d_curData,
					&CV_IMAGE_ELEM(h_img, unsigned char, i, 0),
					channels*h_img->width);
				d_curData += (channels*h_img->width);
			}
		}
		else
		{
			for(int i = 0; i < h_img->height; ++i)
			{
				memcpy(
					&CV_IMAGE_ELEM(h_img, unsigned char, i, 0),
					d_curData,
					channels*h_img->width);
				d_curData += (channels*h_img->width);
			}
		}
	}
}

int main(int argc, char** argv)
{
	int key = -1;
	float fTotalTime = 0;
	unsigned int iFrameCnt, iFrameStart;
	iFrameStart = iFrameCnt = 0;
	
	CvCapture* capture = 0;
	capture = cvCaptureFromAVI(argv[1]);;
	if( !capture )
	{
		fprintf(stderr,"Could not initialize...\n");
		return -1;
	}

	IplImage* videoFrame = NULL;
	CvFastBgGMM* h_pGMMRet = new CvFastBgGMM();
	CvFastBgGMMData* h_pGMMData = new CvFastBgGMMData;
	
	
      
	videoFrame = cvQueryFrame(capture);
	

	if(!videoFrame)
	{
		printf("Bad frame \n");
		exit(0);
	}

	cvNamedWindow("BG", 1);
	cvNamedWindow("FG", 1);
	cvNamedWindow("In Card", 1);
	
// 	CvFastBgGMMParams* pGMMParams = 0;
// 	CvFastBgGMM* pGMM = NULL;
	//creating device pointers
	
	//pGMMParams = cvCreateFastBgGMMParams(videoFrame->width, videoFrame->height);
	
	//create parameters
	
	int width = videoFrame->width; 
	int height = videoFrame->height;

	// setting GMMParams space
	CvFastBgGMMParams* pGMMParams=new CvFastBgGMMParams();
	cutilSafeCall(cudaSetDeviceFlags(cudaDeviceMapHost));
	
	
	

	// setting GMMData space
	
	

	
	int size = width*height;
	pGMMParams->nWidth = width;
	pGMMParams->nHeight = height;
	pGMMParams->nSize = size;

	pGMMParams->nNBands=3;	//always 3 - not implemented for other values!

	//set parameters
	// K - max number of Gaussians per pixeld_GMM
	pGMMParams->nM = 4;			
	// Tb - the threshold - n var
	pGMMParams->fTb = 4*4;
	// Tbf - the threshold
	pGMMParams->fTB = 0.9f;//1-cf from the paper 
	// Tgenerate - the threshold
	pGMMParams->fTg = 3.0f*3.0f;//update the mode or generate new
	pGMMParams->fSigma= 11.0f;//sigma for the new mode
	// alpha - the learning factor
	pGMMParams->fAlphaT=0.001f;
	// complexity reduction prior constant
	pGMMParams->fCT=0.05f;

	//shadow
	// Shadow detection
	pGMMParams->bShadowDetection = 1;//turn on
	pGMMParams->fTau = 0.5f;// Tau - shadow threshold

	pGMMParams->bRemoveForeground = 0;
	
	
	// modify other params in pGMMParams
	//pGMMParams->fAlphaT = 0.0001f;
	pGMMParams->fAlphaT = 0.008f;

	// officially create the parameter (on device memory)
	//pGMM = cvCreateFastBgGMM(pGMMParams, videoFrame,d_GMMData,d_arrImageInfo,d_GMMParams);
	
	
	
	
	if(InitCUDA(h_pGMMRet) < 0)
	{
		delete h_pGMMRet;
		return NULL;
	}

		

	// allocate device global memory
	int iElemCount = pGMMParams->nSize * pGMMParams->nM * sizeof(float);
	int iSizeCount = pGMMParams->nSize * sizeof(int);

	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->ucGaussian), 4*iElemCount));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->rWeight), iElemCount));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMData->rnUsedModes), iSizeCount));
	CUDAGMM_SAFE_CALL(cudaMemset(h_pGMMData->rnUsedModes, 0, iSizeCount));
	
// 	CUDAGMM_SAFE_CALL(cudaMemcpyToSymbol("d_GMMData", &h_pGMMData, sizeof(CvFastBgGMMData), 0, cudaMemcpyHostToDevice));

	
	h_pGMMRet->internal_data = h_pGMMData;
	
	// we will use 4-channels image as input data!
	h_pGMMRet->inputFrame = cvCreateImage(cvSize(pGMMParams->nWidth, pGMMParams->nHeight), IPL_DEPTH_8U, 4);
	h_pGMMRet->nInputImgSize = 4 * videoFrame->width *  pGMMParams->nHeight;
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_inputImg), h_pGMMRet->nInputImgSize));
	h_pGMMRet->h_outputImg = cvCreateImage(cvSize(pGMMParams->nWidth, pGMMParams->nHeight), IPL_DEPTH_8U, 1);
	h_pGMMRet->nOutputImgSize = pGMMParams->nWidth *  pGMMParams->nHeight;
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_outputImg), h_pGMMRet->nOutputImgSize));
	
	// d_arrImageInfo constant (device mem.)
	int inpPixelCnt = pGMMParams->nWidth * pGMMParams->nHeight;

	// number of pixels per thread must be 4k, i.e. 4, 8, 12, 16, 20...
	int iPixelsPerThread = (int)ceil(inpPixelCnt *1.0 / (h_pGMMRet->nBlocksPerGrid * h_pGMMRet->nThreadsPerBlock));
	iPixelsPerThread = 4*(int)ceil(iPixelsPerThread/4.0f);
	h_pGMMRet->nBlocksPerGrid = (int)ceil(inpPixelCnt*1.0 / ((h_pGMMRet->nThreadsPerBlock) * iPixelsPerThread));

	printf("%d pixels/thread, %d threads/block, %d blocks\r\n", 
		iPixelsPerThread, h_pGMMRet->nThreadsPerBlock, h_pGMMRet->nBlocksPerGrid);

	int arrImgInfo[ImageInfoCount] = {	inpPixelCnt, iPixelsPerThread };
	
// 	CUDAGMM_SAFE_CALL(cudaMemcpyToSymbol("d_arrImageInfo", arrImgInfo, ImageInfoCount*sizeof(int), cudaMemcpyHostToDevice));
	
	int *d_arrImageInfo;
	CUDAGMM_SAFE_CALL(cudaMalloc((void **)&d_arrImageInfo, ImageInfoCount*sizeof(int)));
	CUDAGMM_SAFE_CALL(cudaMemcpy(d_arrImageInfo, arrImgInfo, ImageInfoCount*sizeof(int), cudaMemcpyHostToDevice));
	printf("allocated d_array \n");
	CUDAGMM_SAFE_CALL(cudaStreamCreate(&(h_pGMMRet->copyStream)));
	CUDAGMM_SAFE_CALL(cudaStreamCreate(&(h_pGMMRet->execStream)));
	printf("created streams \n");
	printf("%d %d \n",&(h_pGMMRet->h_pinnedIn),h_pGMMRet->nInputImgSize);
	  
	CUDAGMM_SAFE_CALL(cudaHostAlloc((void**)&(h_pGMMRet->h_pinnedIn), h_pGMMRet->nInputImgSize, cudaHostAllocMapped));
	CUDAGMM_SAFE_CALL(cudaHostAlloc((void**)&(h_pGMMRet->h_pinnedOut), h_pGMMRet->nOutputImgSize, cudaHostAllocMapped));
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_inputImg), h_pGMMRet->nInputImgSize))
	CUDAGMM_SAFE_CALL(cudaMalloc((void**)&(h_pGMMRet->d_outputImg), h_pGMMRet->nOutputImgSize));
	printf("allocated on host and device \n");
	// copy the algorithm parameters to Constant memory
	pGMMParams->fPrune = -(pGMMParams->fAlphaT) * (pGMMParams->fCT);
// 	
	printf("just before malloc \n");
	//CUDAGMM_SAFE_CALL(cudaMalloc((void**)d_GMMParams, sizeof(CvFastBgGMMParams)));
	printf("starting memcopy \n");
// 	CUDAGMM_SAFE_CALL(cudaMemcpy((void*)d_GMMParams, (const void*)pGMMParams, sizeof(CvFastBgGMMParams), cudaMemcpyHostToDevice));
	printf("finished memcopy before the second copy \n");
	// setup the initial state for asynchronous execution
	cvCvtColor(videoFrame, h_pGMMRet->inputFrame, CV_BGR2BGRA);
	copyImageData<true>(h_pGMMRet->inputFrame, h_pGMMRet->h_pinnedIn, 4);
	cudaMemcpyAsync(h_pGMMRet->d_inputImg, h_pGMMRet->h_pinnedIn, h_pGMMRet->nInputImgSize, cudaMemcpyHostToDevice, h_pGMMRet->copyStream);
	cudaStreamSynchronize(h_pGMMRet->copyStream);

	cudaUpdateFastBgGMM_Wrapper
		( h_pGMMRet->d_inputImg, h_pGMMRet->d_outputImg,(h_pGMMRet->nBlocksPerGrid),(h_pGMMRet->nThreadsPerBlock),4,h_pGMMRet->execStream,
		  d_arrImageInfo,*h_pGMMData,*pGMMParams
		);
	CUDAGMM_SAFE_CALL( cudaThreadSynchronize() );

	cudaMemcpyAsync(h_pGMMRet->h_pinnedOut, h_pGMMRet->d_outputImg, h_pGMMRet->nOutputImgSize, cudaMemcpyDeviceToHost, h_pGMMRet->copyStream);
	cudaStreamSynchronize(h_pGMMRet->copyStream);
	copyImageData<false>(h_pGMMRet->h_outputImg, h_pGMMRet->h_pinnedOut, 1);
	
 printf("kernel ran once \n");

while(key != 1)
	{

		videoFrame = cvQueryFrame(capture);
		if(!videoFrame)
	{
		printf("Bad frame \n");
		exit(0);
	}
		

		iFrameCnt++;
		double fEllapsed;



		// Update model

// 		fEllapsed = cvUpdateFastBgGMMTimer(pGMM, videoFrame,d_arrImageInfo,d_GMMData,d_GMMParams);
		//run back ground sustraction
// 		cvUpdateFastBgGMM(pGMM, inputImg,d_arrImageInfo,d_GMMData,d_GMMParams);
	cvCvtColor(videoFrame, h_pGMMRet->inputFrame, CV_BGR2BGRA);
	copyImageData<true>(h_pGMMRet->inputFrame, h_pGMMRet->h_pinnedIn, 4);
	cudaMemcpyAsync(h_pGMMRet->d_inputImg, h_pGMMRet->h_pinnedIn, h_pGMMRet->nInputImgSize, cudaMemcpyHostToDevice, h_pGMMRet->copyStream);
	cudaStreamSynchronize(h_pGMMRet->copyStream);

	cudaUpdateFastBgGMM_Wrapper
		( h_pGMMRet->d_inputImg, h_pGMMRet->d_outputImg,(h_pGMMRet->nBlocksPerGrid),(h_pGMMRet->nThreadsPerBlock),4,h_pGMMRet->execStream,d_arrImageInfo,*h_pGMMData,*pGMMParams
		);
	CUDAGMM_SAFE_CALL( cudaThreadSynchronize() );

	cudaMemcpyAsync(h_pGMMRet->h_pinnedOut, h_pGMMRet->d_outputImg, h_pGMMRet->nOutputImgSize, cudaMemcpyDeviceToHost, h_pGMMRet->copyStream);
	cudaStreamSynchronize(h_pGMMRet->copyStream);
	copyImageData<false>(h_pGMMRet->h_outputImg, h_pGMMRet->h_pinnedOut, 1);
#ifdef _DEBUG
	cudaError_t error = cudaGetLastError();
	if(error != cudaSuccess)
	{
		printf("CUDA error: %d: %s\r\n", error, cudaGetErrorString(error));
	}
#endif


		

// 		char sOutput[INT_MAX];
		//sprintf_s(sOutput, "K:\\Ret\\gpu\\ret%05d.png", iFrameCnt);
  		cvSaveImage("image.png", h_pGMMRet->h_outputImg);
		cvSaveImage("BG.png", videoFrame);
		cvSaveImage("In Card.png", h_pGMMRet->inputFrame);
		//sprintf_s(sOutput, "K:\\Ret\\original\\ret%05d.png", iFrameCnt);
// 		cvSaveImage(sOutput, videoFrame);

		cvShowImage("BG", videoFrame);
		cvShowImage("In Card", h_pGMMRet->inputFrame);
		cvShowImage("FG", h_pGMMRet->h_outputImg);	
		
		key = cvWaitKey(10);
		
	}

	iFrameCnt -= iFrameStart;
// 	printf("Average %.1f ms/frame, %.1f FPS\r\n", fTotalTime / (float)iFrameCnt, 1000.0f*(float)iFrameCnt/fTotalTime);
	int tst;
	scanf ( "%d", tst);
	cvDestroyWindow("BG");
	cvDestroyWindow("FG");
	cvDestroyWindow("In Card");
	//free memory    
	//cvReleaseFastBgGMM(&pGMM);
	
		CvFastBgGMM* h_pGMM = h_pGMMRet;

	cvReleaseImage(&(h_pGMM->h_outputImg));
	cvReleaseImage(&(h_pGMM->inputFrame));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_inputImg));
	CUDAGMM_SAFE_CALL( cudaFree(h_pGMM->d_outputImg));
	CUDAGMM_SAFE_CALL( cudaFreeHost(h_pGMM->h_pinnedIn));
	CUDAGMM_SAFE_CALL( cudaFreeHost(h_pGMM->h_pinnedOut));
	CUDAGMM_SAFE_CALL(cudaStreamDestroy(h_pGMM->copyStream));

	CvFastBgGMMData* h_pGMMData2 = h_pGMM->internal_data;
	CUDAGMM_SAFE_CALL(cudaFree(h_pGMMData2->ucGaussian));
	CUDAGMM_SAFE_CALL(cudaFree(h_pGMMData2->rWeight));
	CUDAGMM_SAFE_CALL(cudaFree(h_pGMMData2->rnUsedModes));

	delete h_pGMM->internal_data;
	delete h_pGMM;
	
	
	delete pGMMParams;
	cvReleaseCapture(&capture);

	return 0;
}
